import warnings

from pybasilica.run import fit
from pybasilica.simulation import generate_model

warnings.simplefilter('always', DeprecationWarning)

warnings.warn(
    "The 'pybasilica' package is deprecated and will no longer be maintained."
    "Please switch to 'pybascule' for future updates.",
    DeprecationWarning
)



# from matplotlib import interactive
from rich.console import Console
from rich.table import Table
from rich import box
import pandas as pd
import numpy as np
from rich.progress import Progress, BarColumn, TextColumn, TaskProgressColumn, TimeRemainingColumn, SpinnerColumn, RenderableColumn
from rich.live import Live
from rich.table import Table
from sys import maxsize
from copy import deepcopy

from pybasilica.svi import PyBasilica
from pybasilica.svi_mixture import PyBasilica_mixture


def fit(x=None, alpha=None, k_list=[0,1,2,3,4,5], lr = 0.005, optim_gamma = 0.1, n_steps = 500, enumer = "parallel", 
        cluster = None, beta_fixed = None, hyperparameters = None, dirichlet_prior = True, autoguide = False,
        compile_model = False, CUDA = False, nonparametric = False, stage = "",  seed_list = [10], 
        store_parameters = False, store_fits=False):

    if isinstance(seed_list, int): seed_list = [seed_list]
    if isinstance(cluster, int) and cluster < 1: cluster = None
    elif isinstance(cluster, int): cluster = [cluster]
    if isinstance(k_list, int): k_list = [k_list]

    if x is None and alpha is None: raise "Both count and exposure matrices are None."

    kwargs = {
        "x":x,
        "lr":lr,
        "optim_gamma":optim_gamma,
        "n_steps":n_steps,
        "enumer":enumer,
        "dirichlet_prior":dirichlet_prior,
        "beta_fixed":beta_fixed,
        "hyperparameters":hyperparameters,
        "compile_model":compile_model,
        "CUDA":CUDA,
        "store_parameters":store_parameters,
        "stage":stage,
        }

    kwargs_mixture = {
        "lr":lr,
        "optim_gamma":optim_gamma,
        "n_steps":n_steps,
        "enumer":enumer,
        "hyperparameters":hyperparameters,
        "autoguide":autoguide,
        "compile_model":compile_model,
        "CUDA":CUDA,
        "store_parameters":store_parameters,
        "nonparam":nonparametric
    }

    if nonparametric and isinstance(cluster, list): cluster = [max(cluster)]

    bestK, scoresK, fitsK, fits_alpha = None, None, dict(), dict()
    if x is not None:
        bestK, scoresK, fitsK = run_fit(seed_list=seed_list, kwargs=kwargs, parname="k_denovo",
                                        parlist=k_list, score_name="bic", store_fits=store_fits, cls=PyBasilica)
        bestK.scores = scoresK
        bestK.fits = fitsK

    if cluster is not None:
        if bestK is not None: alpha = bestK.params["alpha"]

        kwargs_mixture["alpha"] = alpha
        bestCL, scoresCL, fitsCL = run_fit(seed_list=seed_list, kwargs=kwargs_mixture, parname="cluster",
                                           parlist=list(cluster), score_name="icl", store_fits=store_fits,
                                           cls=PyBasilica_mixture)
        bestCL.scores = scoresCL
        bestCL.fits = fitsCL

        bestK = merge_k_cl(obj=bestK, obj_mixt=bestCL, store_parameters=store_parameters, store_fits=store_fits) if bestK is not None else bestCL

    if bestK is not None: bestK.convert_to_dataframe(x) if x is not None else bestK.convert_to_dataframe(alpha)
    if store_fits: convert_fits(bestK)

    return bestK


def single_run(seed_list, kwargs, cls, score_name, idd):
    best_score, best_run = maxsize, None
    fits, scores = dict(), dict()

    for seed in seed_list:
        idd_s = "seed:"+str(seed)
        obj = cls(seed=seed, **kwargs)
        obj._fit()
        obj.idd = idd + "." + idd_s

        sc_s = obj.__dict__[score_name]

        if best_run is None or sc_s < best_score:
            best_score = sc_s
            best_run = deepcopy(obj)

        scores[idd_s] = {"bic":obj.bic, "aic":obj.aic, "icl":obj.icl, "llik":obj.likelihood}
        fits[idd_s] = obj

    return best_run, scores, fits


def run_fit(seed_list, kwargs, parname, parlist, score_name, store_fits, cls):
    '''
    `seed_list` -> list of seeds to test \\
    `kwargs` -> dict of arguments for the fit \\
    `parname` -> name of the parameter in the for loop \\
    `parlist` -> list of values to iterate through \\
    `score_name` -> name of the score to minimize (either `bic`, `icl` or `aic`) \\
    `store_fits` -> if True, all fits will be stores \\
    `cls` -> class to be used for the fit \\
    '''
    best_score, best_run = maxsize, None
    fits, scores = dict(), dict()
    for i in parlist:
        idd_i = parname + ":" + str(i)

        # fits_i contains a dict with the fits for all seeds in seed_list
        kwargs[parname] = i
        best_i, scores_i, fits_i = single_run(seed_list=seed_list, kwargs=kwargs, cls=cls, score_name=score_name, idd=idd_i)

        if parname == "cluster": idd_i = parname + ":" + str(i) + "_" + str(len(np.unique(best_i.groups)))

        sc_i = best_i.__dict__[score_name]

        if best_run is None or sc_i < best_score:
            best_score = sc_i
            best_run = deepcopy(best_i)

        scores[idd_i] = scores_i
        if store_fits: 
            fits[idd_i] = fits_i

    return best_run, scores, fits


def merge_k_cl(obj, obj_mixt, store_parameters, store_fits):
    obj.__dict__["fits"] = {"NMF":obj.fits, "CL":obj_mixt.fits}
    obj.__dict__["scores"] = {"NMF":obj.scores, "CL":obj_mixt.scores}
    obj.__dict__["idd"] = obj.idd + ";" + obj_mixt.idd

    obj.gradient_norms = {**obj.gradient_norms, **obj_mixt.gradient_norms}
    if store_parameters:
        obj.train_params = [{**obj.train_params[i], **obj_mixt.train_params[i]} for i in range(len(obj.train_params))]
    obj.losses_dmm = obj_mixt.losses
    obj.likelihoods_dmm = obj_mixt.likelihoods
    obj.regs_dmm = obj_mixt.regs
    obj.groups, obj.n_groups = obj_mixt.groups, obj_mixt.n_groups
    obj.params = {**obj.params, **obj_mixt.params}
    obj.init_params = {**obj.init_params, **obj_mixt.init_params}
    obj.hyperparameters = {**obj.hyperparameters, **obj_mixt.hyperparameters}
    return obj


def convert_fits(obj):
    try: _convert_fits_aux(obj.fits, input=obj.x)
    except: pass
    try: _convert_fits_aux(obj.fits, input=obj.alpha)
    except: pass
    try: _convert_fits_aux(obj.fits["NMF"], input=obj.x)
    except: pass
    try: _convert_fits_aux(obj.fits["CL"], input=obj.params["alpha"])
    except: pass


def _convert_fits_aux(fits, input):
    for _, v_1 in fits.items(): 
        for _, v_2 in v_1.items():
            v_2.convert_to_dataframe(input)




from logging import raiseExceptions
import numpy as np
import torch
import pyro.distributions as dist
import torch.nn.functional as F

#========================================== OOP ========================



#========================================== OOP ========================


#-----------------------------------------------------------------[PASSESD]
def get_alpha_beta(params):
    alpha = torch.exp(params["alpha"])
    alpha = alpha / (torch.sum(alpha, 1).unsqueeze(-1))

    if params["k_denovo"] > 0:
        beta = torch.exp(params["beta"])
        beta = beta / (torch.sum(beta, 1).unsqueeze(-1))
    else:
        beta = None
    return  alpha, beta
    # alpha : torch.Tensor (num_samples X  k)
    # beta  : torch.Tensor (k_denovo    X  96) | ZERO ( 0 )

#-----------------------------------------------------------------[PASSED]
def compute_bic(params):
    alpha, beta_denovo = get_alpha_beta(params)

    if beta_denovo is None:
        beta = params["beta_fixed"]
        k_fixed = params["beta_fixed"].shape[0]
    elif params["beta_fixed"] is None:
        beta = beta_denovo
        k_fixed = 0
    else:
        beta = torch.cat((params["beta_fixed"], beta_denovo), axis=0)
        k_fixed = params["beta_fixed"].shape[0]

    theta = torch.sum(params["M"], axis=1)

    log_L_Matrix = dist.Poisson(
        torch.matmul(
            torch.matmul(torch.diag(theta), alpha), 
            beta)
            ).log_prob(params["M"])
    log_L = torch.sum(log_L_Matrix)
    log_L = float("{:.3f}".format(log_L.item()))

    k = (params["M"].shape[0] * (k_fixed + params["k_denovo"])) + (params["k_denovo"] * params["M"].shape[1])
    n = params["M"].shape[0] * params["M"].shape[1]
    bic = k * torch.log(torch.tensor(n)) - (2 * log_L)
    return bic.item()



#-----------------------------------------------------------------[PASSED]
def fixedFilter(alpha_tensor, beta_df, theta_np, fixedLimit):
    # alpha_tensor (inferred alpha) ------------------------- dtype: torch.Tensor
    # beta_df (input beta fixed) ---------------------------- dtype: data.frame
    # theta_np ---------------------------------------------- dtype: numpy

    if beta_df is None:
        return []

    beta_test_list = list(beta_df.index)
    #k_fixed = len(beta_test_list)

    alpha = np.array(alpha_tensor)
    theta = np.expand_dims(theta_np, axis = 1)
    x = np.multiply(alpha, theta)
    a = np.sum(x, axis=0) / np.sum(x)
    b = [x for x in a if x <= fixedLimit]

    #a = (torch.sum(alpha_inf, axis=0) / np.array(alpha_inf).shape[0]).tolist()[:k_fixed]
    #b = [x for x in a if x <= 0.05]

    a = list(a)
    b = list(b)
    
    excluded = []
    if len(b)==0:
        return beta_test_list   # all signatures are significant
    else:
        for j in b:
            index = a.index(j)
            if index < len(beta_test_list):
                excluded.append(beta_test_list[index])
            #print("Signature", beta_test_list[index], "is not included!")
    
    for k in excluded:
        beta_test_list.remove(k)
    
    # list of significant signatures in beta_test ---- dtype: list
    return beta_test_list  # dtype: list

#-----------------------------------------------------------------[PASSED]
def denovoFilter(beta_inferred, cosmic_df, delta):
    # beta_inferred -- dtype: tensor
    # cosmic_df ------ dtype: dataframe
    # delta ---------- dtype: float

    match_list = []
    
    if beta_inferred is None:
        return match_list

    for denovo_index, denovo_row in enumerate(beta_inferred):
        denovo = denovo_row[None, :]    # dtype: tensor (convert from 1D to 2D)
        max_score = 0
        match = ""

        for reference_name, reference_row in cosmic_df.iterrows():
            reference = torch.tensor(reference_row.values).float()    # dtype: tensor
            reference = reference[None, :]                      # dtype: tensor (convert from 1D to 2D)

            #score = F.kl_div(denovo, cos_tensor, reduction="batchmean").item()
            score = F.cosine_similarity(denovo, reference).item()
            if score >= max_score:
                max_score = score
                match = reference_name

        if max_score > delta:
            match_list.append(match)
        
    return match_list    # dtype: list & torch.tensor


def stopRun(new_list, old_list, denovo_list):
    # new_list      dtype: list
    # old_list      dtype: list
    # denovo_list   dtype: list
    new_list.sort()
    old_list.sort()
    if new_list==old_list and len(denovo_list)==0:
        return True
    else:
        return False


def initialize_params(M, groups, B_input, lr, steps):

    params = {}

    params["M"] = torch.tensor(M.values).float()
    if B_input is None:
        params["beta_fixed"] = None
    else:
        params["beta_fixed"] = torch.tensor(B_input.values).float()
    params["lr"] = lr
    params["steps_per_iter"] = int(steps)
    params["groups"] = groups

    return params


#------------------------ DONE! ----------------------------------[passed]
# note: just check the order of kl-divergence arguments and why the value is negative
def regularizer(beta_fixed, beta_denovo):
    loss = 0
    for fixed in beta_fixed:
        for denovo in beta_denovo:
            loss += F.kl_div(fixed, denovo, reduction="batchmean").item()

    return loss

#------------------------ DONE! ----------------------------------[passed]
def custom_likelihood(M, alpha, beta_fixed, beta_denovo):
    # build full signature profile (beta) matrix

    if beta_fixed is None and beta_denovo is None:
        raise Exception("wrong input!")

    elif beta_fixed is None:
        #print("beta_fixed is None")
        beta = beta_denovo
        regularization = 0

    elif beta_denovo is None:
        #print("beta_denovo is None")
        beta = beta_fixed
        regularization = 0

    else:
        beta = torch.cat((beta_fixed, beta_denovo), axis=0)
        regularization = regularizer(beta_fixed, beta_denovo)
    
    likelihood =  dist.Poisson(torch.matmul(torch.matmul(torch.diag(torch.sum(M, axis=1)), alpha), beta)).log_prob(M)
    
    return likelihood + regularization




import numpy as np
import pandas as pd
import torch
import pyro
import pyro.distributions.constraints as constraints
import pyro.distributions as dist
import torch.nn.functional as F
from pyro.infer import SVI, Trace_ELBO, JitTrace_ELBO
from collections import defaultdict
from tqdm import trange


MIN_POS_N = torch.finfo(torch.float32).tiny


class PyBasilica():
    def __init__(
        self,
        x,
        k_denovo,
        n_steps,
        lr = 0.001,
        optim_gamma = 0.1,
        enumer = "parallel",
        hyperparameters = None,
        dirichlet_prior = True,
        beta_fixed = None,
        compile_model = True,
        CUDA = False,
        enforce_sparsity = True,
        store_parameters = False,

        stage = "", 
        seed = 10
        ):

        self._set_data_catalogue(x)
        self._hyperpars_default = {"alpha_sigma":0.1, "alpha_p_conc0":0.6, "alpha_conc":1., 
                                   "omega_conc":10., "eps_sigma":10, "pi_conc0":0.6, 
                                   "penalty_scale":0}
        self._set_fit_settings(enforce_sparsity=enforce_sparsity, lr=lr, optim_gamma=optim_gamma, n_steps=n_steps,
                               compile_model=compile_model, CUDA=CUDA, store_parameters=store_parameters, 
                               stage=stage, seed=seed)

        self._set_k_denovo(k_denovo)  # set self.k_denovo
        self._set_beta_fixed(beta_fixed)  # set self.beta_fixed and self.k_fixed
        self._fix_zero_denovo_null_reference()
        self.K = self.k_denovo + self.k_fixed
        self.K_alpha = self.K

        self._set_hyperparams(enumer=enumer, hyperparameters=hyperparameters)

        self._init_seed = None


    def _set_fit_settings(self, enforce_sparsity, lr, optim_gamma, n_steps, 
                          compile_model, CUDA, store_parameters, stage, seed):
        self.enforce_sparsity = enforce_sparsity
        self.lr = lr
        self.optim_gamma = optim_gamma
        self.n_steps = int(n_steps)
        self.compile_model = compile_model
        self.CUDA = CUDA

        self.store_parameters = store_parameters
        self.stage = stage

        self.seed = seed


    def _set_hyperparams(self, enumer, hyperparameters):
        self.enumer = enumer
        self.init_params = None

        if hyperparameters is None:
            self.hyperparameters = self._hyperpars_default
        else:
            self.hyperparameters = dict()
            for parname, value in self._hyperpars_default.items():
                self.hyperparameters[parname] = hyperparameters.get(parname, value)

        self.hyperparameters["omega_conc"] = torch.cat((torch.ones(self.k_denovo, self.k_fixed) * self.hyperparameters["omega_conc"], 
                                                        torch.ones(self.k_denovo, 1)), dim=1)

        alpha_conc = self.hyperparameters["alpha_conc"]
        if self.k_fixed > 0:
            if (isinstance(alpha_conc, list)):
                if len(alpha_conc) != self.k_fixed:
                    raise "The length of `alpha_conc` list must be the same as the number of fixed signatures."
                
                self.hyperparameters["alpha_conc"] = torch.cat((torch.tensor(alpha_conc, dtype=torch.float64), 
                                                                torch.ones(self.k_denovo, dtype=torch.float64)))

            elif isinstance(alpha_conc, int) or isinstance(alpha_conc, float):
                # if given as input, the fixed signatures will have a user-defined concentration
                self.hyperparameters["alpha_conc"] = torch.cat((alpha_conc * torch.ones(self.k_fixed, dtype=torch.float64),
                                                                torch.ones(self.k_denovo, dtype=torch.float64)))
            else:
                raise "`alpha_conc` must be either a list or a number."
        
        else:
            self.hyperparameters["alpha_conc"] = self._hyperpars_default["alpha_conc"] * torch.ones(self.k_denovo, dtype=torch.float64)

        for k, v in self.hyperparameters.items():
            self.hyperparameters[k] = self._to_gpu(v)


    def _fix_zero_denovo_null_reference(self):
        '''
        If there are zero denovo (k=0) and the reference is empty the model will only fit a random noise.
        '''
        if self.k_denovo == 0 and self.k_fixed == 0:
            self.stage = "random_noise"
            self.beta_fixed = torch.zeros(1, self.contexts, dtype=torch.float64)
            self.k_fixed = 1
            self._noise_only = True
        else:
            self._noise_only = False


    def _set_data_catalogue(self, x):
        try:
            self.x = torch.tensor(x.values, dtype=torch.float64)
            self.n_samples = x.shape[0]
            self.contexts = x.shape[1]
        except:
            raise Exception("Invalid mutations catalogue, expected Dataframe!")


    def _set_beta_fixed(self, beta_fixed):
        if beta_fixed is None:
            self.beta_fixed, self.k_fixed, self.fixed_names = None, 0, []
            return

        self.alpha_conc = None
        self.fixed_names = list(beta_fixed.index) if isinstance(beta_fixed, pd.DataFrame) else ["F"+str(f+1) for f in range(beta_fixed.shape[0])]

        if isinstance(beta_fixed, pd.DataFrame):
            beta_fixed = beta_fixed.values

        self.beta_fixed = torch.tensor(beta_fixed, dtype=torch.float64)
        if len(self.beta_fixed.shape) == 1: self.beta_fixed = self.beta_fixed.reshape(1, self.beta_fixed.shape[0])

        self.k_fixed = beta_fixed.shape[0]
        if self.k_fixed > 0: self._fix_zero_contexts()

    def _fix_zero_contexts(self):
        '''
        If a context has a density of 0 in all signatures but X contains mutations in that context, an error will raise.
        A value of 1e-07 is set to one random signature in such contexts.
        '''
        colsums = torch.sum(self.beta_fixed, axis=0)
        zero_contexts = torch.where(colsums==0)[0]
        if torch.any(colsums == 0):
            random_sig = [0] if self.k_fixed == 1 else torch.randperm(self.beta_fixed.shape[0])[:torch.numel(zero_contexts)]

            for rr in random_sig:
                self.beta_fixed[rr, zero_contexts] = 1e-07

            self.beta_fixed = self._norm_and_clamp(self.beta_fixed)


    def _set_k_denovo(self, k_denovo):
        if isinstance(k_denovo, int):
            self.k_denovo = k_denovo
        else:
            raise Exception("Invalid k_denovo value, expected integer!")


    def model(self):
        n_samples, n_contexts = self.n_samples, self.contexts

        if self._noise_only: alpha = self._to_gpu(torch.zeros(self.n_samples, 1, dtype=torch.float64))
        # Flat model
        if not self._noise_only:
            with pyro.plate("n", self.n_samples):
                alpha = pyro.sample("latent_exposure", dist.Dirichlet(self.hyperparameters["alpha_conc"]))

        epsilon = None
        if self.stage == "random_noise":
            with pyro.plate("contexts3", self.contexts):  # columns
                with pyro.plate("n3", n_samples):  # rows
                    epsilon = pyro.sample("latent_m", dist.HalfNormal(self.hyperparameters["eps_sigma"]))

        # Beta
        beta_denovo = None
        if self.k_denovo > 0:
            ## directly learn the beta_denovo
            beta_weight = None
            beta_conc = self._get_beta_centroid(eps=1e-15, power=1)
            with pyro.plate("k_denovo", self.k_denovo):  # rows
                beta_denovo = pyro.sample("latent_signatures", dist.Dirichlet(beta_conc))

            beta = self._get_unique_beta(beta_fixed=self.beta_fixed, 
                                         beta_denovo=beta_denovo, 
                                         beta_weights=beta_weight)

        else: 
            beta = self._get_unique_beta(beta_fixed=self.beta_fixed,
                                         beta_denovo=beta_denovo)

        a = torch.matmul(torch.matmul(torch.diag(torch.sum(self.x, axis=1)), alpha), beta)
        if self.stage == "random_noise": a = a + epsilon
        with pyro.plate("contexts", n_contexts):
            with pyro.plate("n2", n_samples):
                pyro.sample("obs", dist.Poisson(a), obs=self.x)

        if self.k_denovo > 0:
            alpha_recomputed = self._get_alpha_stick_breaking(alpha_star=alpha, beta_weights=beta_weight)
            beta_fixed_cum = self._compute_cum_beta_fixed(self.beta_fixed)
            penalty = self._compute_penalty(alpha=alpha_recomputed, beta_fixed_cum=beta_fixed_cum, beta_denovo=beta_denovo)
            pyro.factor("penalty", penalty)


    def _get_beta_centroid(self, eps=1e-10, power=2):
        beta_fixed_cum = self._compute_cum_beta_fixed(self.beta_fixed)
        max_val = beta_fixed_cum.max()
        centroid = (max_val - beta_fixed_cum + eps) ** power
        return centroid * 100


    def _compute_cum_beta_fixed(self, beta_fixed):
        if self.beta_fixed is not None:
            return torch.sum(beta_fixed, dim=0)
        return dist.Dirichlet(torch.ones(self.contexts).double()).sample()


    def _compute_penalty(self, alpha, beta_fixed_cum, beta_denovo):
        alpha_fixed = torch.sum(alpha[:,:self.k_fixed], dim=0).unsqueeze(1)
        alpha_denovo = torch.sum(alpha[:,self.k_fixed:], dim=0).unsqueeze(1)

        w_fixed = torch.sum(beta_fixed_cum * alpha_fixed, dim=0)
        w_denovo = torch.sum(beta_denovo * alpha_denovo, dim=0)

        w_difference = torch.abs(w_fixed - w_denovo)
        res = torch.sum(w_difference)

        return res * self.hyperparameters["penalty_scale"]


    def guide(self):
        n_samples, k_denovo = self.n_samples, self.k_denovo
        init_params = self._initialize_params()

        if not self._noise_only:
            with pyro.plate("n", n_samples):
                alpha = pyro.param("alpha", init_params["alpha"], constraint=constraints.simplex)
                pyro.sample("latent_exposure", dist.Delta(alpha).to_event(1))

        # Epsilon 
        if self.stage == "random_noise":
            eps_sigma = pyro.param("lambda_epsilon", init_params["epsilon_var"], constraint=constraints.greater_than(MIN_POS_N))

            with pyro.plate("contexts3", self.contexts):
                with pyro.plate("n3", n_samples):
                    pyro.sample("latent_m", dist.HalfNormal(eps_sigma))

        # Beta
        if k_denovo > 0:
            beta_param = pyro.param("beta_denovo", lambda: init_params["beta_dn_param"], constraint=constraints.simplex)
            with pyro.plate("k_denovo", k_denovo):
                pyro.sample("latent_signatures", dist.Delta(beta_param).to_event(1))


    def _initialize_params_nonhier(self):
        pi = alpha_prior = alpha = epsilon = beta_dn = beta_weights = pi_conc0 = None

        eps_sigma = self.hyperparameters["eps_sigma"]
        pi_conc0 = dist.Gamma(0.001, 0.001).sample([self.k_fixed])
        alpha = dist.Dirichlet(self.hyperparameters["alpha_conc"]).sample((self.n_samples,))

        if self.stage == "random_noise":
            epsilon = dist.HalfNormal(torch.ones(self.n_samples, self.contexts, dtype=torch.float64) * eps_sigma).sample()

        if self.k_denovo > 0:
            beta_conc = self._get_beta_centroid(eps=1e-3, power=1)
            beta_dn = dist.Dirichlet(beta_conc).sample((self.k_denovo,))
            omega_conc = self.hyperparameters["omega_conc"][0]
            # beta_weights = dist.Dirichlet(omega_conc).sample((self.k_denovo,))

        params = self._create_init_params_dict(pi=pi, alpha_prior=alpha_prior, alpha=alpha, epsilon=epsilon, 
                                               beta_dn=beta_dn, beta_weights=beta_weights, pi_conc0=pi_conc0)
        return params


    def _create_init_params_dict(self, pi=None, alpha_prior=None, alpha=None, epsilon=None, 
                                 beta_dn=None, beta_weights=None, pi_conc0=None):
        params = dict()

        if pi is not None: params["pi_param"] = pi
        if alpha_prior is not None: params["alpha_prior_param"] = alpha_prior
        if alpha is not None: params["alpha"] = alpha
        if epsilon is not None: params["epsilon_var"] = epsilon
        if beta_dn is not None: params["beta_dn_param"] = beta_dn
        if beta_weights is not None: params["beta_weights"] = beta_weights
        if pi_conc0 is not None: params["pi_conc0_param"] = pi_conc0

        return params


    def _initialize_params(self):
        if self.init_params is None:
            self.init_params = self._initialize_params_nonhier()

        return self.init_params


    def _get_unique_beta(self, beta_fixed, beta_denovo, beta_weights=None, convert=False):
        if beta_weights is not None:
            return self._get_unique_beta_stick_breaking(beta_fixed, beta_denovo, beta_weights, convert=convert)
        if beta_fixed is None: return beta_denovo
        if beta_denovo is None or self._noise_only: return beta_fixed

        return torch.cat((beta_fixed, beta_denovo), axis=0)


    def _get_unique_beta_stick_breaking(self, beta_fixed, beta_denovo, beta_weights, convert=False):
        beta = torch.zeros(beta_weights.shape[0], self.contexts, dtype=torch.float64)

        for j in range(beta_weights.shape[0]):
            for r in range(self.k_fixed):
                beta[j] += beta_weights[j,r] * beta_fixed[r]
                beta[j] += beta_denovo[j]

        if not convert: return beta
        return np.array(beta)


    def _get_alpha_stick_breaking(self, alpha_star, beta_weights, convert=False):
        if beta_weights is None: 
            if not convert: return alpha_star
            return np.array(alpha_star)
        
        alpha = torch.zeros((self.n_samples, self.K))

        if not isinstance(alpha_star, torch.Tensor): alpha_star = torch.tensor(alpha_star)
        if not isinstance(beta_weights, torch.Tensor): beta_weights = torch.tensor(beta_weights)

        for n in range(self.n_samples):
            alpha_n = alpha_star[n,:]
            for j in range(beta_weights.shape[1]-1):
                beta_w_j = beta_weights[:,j]
                alpha[n,j] = torch.sum(alpha_n * beta_w_j)
            
            alpha[n,j+1:] = beta_weights[:,-1] * alpha_n
        
        if not convert: return self._norm_and_clamp(alpha)
        return np.array(self._norm_and_clamp(alpha))


    def _get_param(self, param_name, normalize=False, to_cpu=True, convert=False):
        try:
            if param_name == "beta_fixed": 
                par = self.beta_fixed
            elif param_name == "alpha" and self._noise_only:
                return self._to_gpu(torch.zeros(self.n_samples, 1, dtype=torch.float64), move=not to_cpu)
            else:
                par = pyro.param(param_name)

            par = self._to_cpu(par, move=to_cpu)
            if isinstance(par, torch.Tensor): par = par.clone().detach()
            if normalize: par = self._norm_and_clamp(par)

            if par is not None and convert:
                par = self._to_cpu(par, move=True)
                par = np.array(par)

            return par

        except Exception as e: return None


    def get_param_dict(self, convert=False, to_cpu=True, all=False):
        params = dict()
        params["alpha"] = self._get_param("alpha", normalize=True, convert=convert, to_cpu=to_cpu)
        params["beta_d"] =  self._get_param("beta_denovo", normalize=True, convert=convert, to_cpu=to_cpu)
        params["beta_f"] = self._get_param("beta_fixed", convert=convert, to_cpu=to_cpu)
        params["beta_w"] = self._get_param("beta_weights", convert=convert, to_cpu=to_cpu, normalize=False)
        params["pi_conc0"] = self._get_param("pi_conc0", normalize=False, convert=convert, to_cpu=to_cpu)

        if self.stage == "random_noise":
            params["lambda_epsilon"] = self._get_param("lambda_epsilon", normalize=False, convert=convert, to_cpu=to_cpu)

        return params


    def _fit(self, set_attributes=True):
        pyro.clear_param_store()  # always clear the store before the inference

        self.x = self._to_gpu(self.x)
        self.beta_fixed = self._to_gpu(self.beta_fixed)

        if self.CUDA and torch.cuda.is_available(): 
            torch.set_default_tensor_type('torch.cuda.FloatTensor')
        else:
            torch.set_default_tensor_type(t=torch.FloatTensor)

        if self.compile_model and not self.CUDA: elbo = JitTrace_ELBO()
        else: elbo = Trace_ELBO()

        lrd = self.optim_gamma ** (1 / self.n_steps)
        optimizer = pyro.optim.ClippedAdam({"lr": self.lr, "lrd":lrd, "clip_norm":1e10})

        pyro.set_rng_seed(self.seed)
        pyro.get_param_store().clear()

        self._initialize_params()

        svi = SVI(self.model, self.guide, optimizer, loss=elbo)
        loss = float(svi.step())

        train_params_each = 2 if self.n_steps <= 100 else int(self.n_steps / 100)
        gradient_norms = defaultdict(list)
        for name, value in pyro.get_param_store().named_parameters():
            value.register_hook(lambda g, name=name: gradient_norms[name].append(g.norm().item()))

        self.losses, self.regs, self.likelihoods, self.train_params = list(), list(), list(), list()
        t = trange(self.n_steps, desc="Bar desc", leave=True)
        for i in t:  # inference - do gradient steps
            loss = float(svi.step())
            self.losses.append(loss)

            self.regs.append(self.compute_regularization(beta_denovo=self._get_param("beta_denovo", to_cpu=False), to_cpu=False))
            self.likelihoods.append(self._likelihood(to_cpu=False))

            if self.store_parameters and i%train_params_each==0: 
                self.train_params.append(self.get_param_dict(to_cpu=True, convert=True, all=False))

            if i%5 == 0:
                t.set_description("ELBO %f" % loss)
                t.refresh()

        if set_attributes is False: return

        self.x = self._to_cpu(self.x)
        self.beta_fixed = self._to_cpu(self.beta_fixed)

        self.gradient_norms = dict(gradient_norms) if gradient_norms is not None else None
        self._set_params()
        self.likelihood = self._likelihood(to_cpu=True)
        self.likelihood_reg = self._likelihood(to_cpu=True) + self.compute_regularization(self._get_param("beta_denovo", to_cpu=True), to_cpu=True)

        self.set_scores()


    def _likelihood(self, to_cpu=False):
        llik = self._likelihood_flat(to_cpu=to_cpu)
        return llik.sum().item()


    def _likelihood_flat(self, to_cpu=False):
        params = self.get_param_dict(convert=False, to_cpu=to_cpu)
        beta = self._get_unique_beta(beta_fixed=params["beta_f"], beta_denovo=params["beta_d"],
                                     beta_weights=params["beta_w"])
        M = self._to_cpu(self.x, move=to_cpu)

        alpha_hat = torch.sum(M, axis=1).unsqueeze(1) * params["alpha"]
        rate = torch.matmul(alpha_hat, beta)

        if self.stage == "random_noise": rate += dist.HalfNormal(params["lambda_epsilon"]).sample()
        llik = dist.Poisson(rate).log_prob(M)

        return llik


    def _norm_and_clamp(self, par):
        mmin = 0
        if torch.any(par < 0): mmin = torch.min(par, dim=-1)[0].unsqueeze(-1)

        nnum = par - mmin
        par = nnum / torch.sum(nnum, dim=-1).unsqueeze(-1)

        return par


    def _set_params(self):
        self.params = self.get_param_dict(convert=True, all=True)


    def set_scores(self):
        self._set_bic()
        self._set_aic()
        self.icl = None


    def compute_regularization(self, beta_denovo, to_cpu=False):
        if beta_denovo is None: return 0
        if not isinstance(beta_denovo, torch.Tensor): beta_denovo = torch.tensor(beta_denovo)
        beta_fixed_cum = self._compute_cum_beta_fixed(self.beta_fixed)
        alpha_star = self._get_param("alpha", normalize=True, to_cpu=to_cpu)
        beta_weights = self._get_param("beta_weights", to_cpu=to_cpu)
        alpha = self._get_alpha_stick_breaking(alpha_star=alpha_star, beta_weights=beta_weights)
        return self._compute_penalty(alpha=alpha, beta_fixed_cum=beta_fixed_cum, beta_denovo=beta_denovo).item()


    def _set_bic(self):
        _log_like = self.likelihood_reg
        k = self._number_of_params() 
        bic = k * torch.log(torch.tensor(self.n_samples, dtype=torch.float64)) - (2 * _log_like)
        self.bic = bic.item()


    def _set_aic(self):
        _log_like = self.likelihood_reg
        k = self._number_of_params() 
        aic = 2*k - 2*_log_like

        if (isinstance(aic, torch.Tensor)): self.aic = aic.item()
        else: self.aic = aic


    def _compute_entropy(self) -> np.array:
        '''
        `entropy(z) = - sum^K( sum^N( z_probs_nk * log(z_probs_nk) ) )`
        `entropy(z) = - sum^K( sum^N( exp(log(z_probs_nk)) * log(z_probs_nk) ) )`
        '''

        matrix = self._compute_denovo_cosine()
        entr = 0
        for n in range(self.n_samples):
            for k in range(self.cluster):
                entr += torch.exp(matrix[k,n]) * matrix[k,n]
        return -entr.detach()


    def _number_of_params(self):
        k = 0
        if self.k_denovo == 0 and self._noise_only: k = 0
        else: k += self.k_denovo * self.contexts # beta denovo

        if self.stage == "random_noise":
            k += self.params["lambda_epsilon"].shape[0] * self.params["lambda_epsilon"].shape[1]  # random noise

        if not self._noise_only: k += self.n_samples * self.K  # alpha if no noise is learned

        return k


    def _to_cpu(self, param, move=True):
        if param is None: return None
        if move and self.CUDA and torch.cuda.is_available() and isinstance(param, torch.Tensor):
            return param.cpu()
        return param


    def _to_gpu(self, param, move=True):
        if param is None: return None
        if move and self.CUDA and torch.cuda.is_available() and isinstance(param, torch.Tensor):
            return param.cuda()
        return param


    def convert_to_dataframe(self, x):
        # mutations catalogue
        self.x = x
        sample_names, contexts = list(x.index), list(x.columns)
        fixed_names = self.fixed_names
        denovo_names = ["D"+str(d+1) for d in range(self.k_denovo)] if self.k_denovo>0 else []

        alpha_columns = fixed_names + denovo_names

        # if self.K_alpha == self.k_denovo: alpha_columns = denovo_names
        # else: alpha_columns = fixed_names + denovo_names

        if self.beta_fixed is not None and isinstance(self.beta_fixed, torch.Tensor) and torch.sum(self.beta_fixed) > 0:
            self.beta_fixed = pd.DataFrame(self.beta_fixed, index=fixed_names, columns=contexts)

        self.params = self._convert_pars(param_dict=self.params, sample_names=sample_names, 
                                         fixed_names=fixed_names, denovo_names=denovo_names, 
                                         contexts=contexts)

        if len(self.train_params) > 0:
            for i,v in enumerate(self.train_params):
                self.train_params[i] = self._convert_pars(param_dict=v, sample_names=sample_names, 
                                                          fixed_names=fixed_names, denovo_names=denovo_names, 
                                                          contexts=contexts)

        for k, v in self.hyperparameters.items():
            if v is None: continue
            v = self._to_cpu(v, move=True)
            if isinstance(v, torch.Tensor): 
                if len(v.shape) == 0: self.hyperparameters[k] = int(v)
                else: self.hyperparameters[k] = v.numpy()

        self._set_init_params(sample_names=sample_names, fixed_names=fixed_names, 
                              denovo_names=denovo_names, contexts=contexts, 
                              alpha_columns=alpha_columns)


    def _convert_pars(self, param_dict, sample_names, fixed_names, denovo_names, contexts):
        for parname, par in param_dict.items():
            if par is None: continue
            par = self._to_cpu(par, move=True)
            if parname == "alpha": 
                param_dict[parname] = pd.DataFrame(np.array(par), index=sample_names, columns=fixed_names+denovo_names)
            elif parname == "beta_d" or parname == "beta_star": 
                param_dict[parname] = pd.DataFrame(np.array(par), index=denovo_names, columns=contexts)
            elif parname == "beta_f": 
                param_dict[parname] = self.beta_fixed
            elif parname == "beta_w":
                param_dict[parname] = pd.DataFrame(np.array(par), index=denovo_names, columns=fixed_names+["DN"])
            elif parname == "pi": 
                param_dict[parname] = par.tolist() if isinstance(par, torch.Tensor) else par
            elif parname == "lambda_epsilon":
                param_dict[parname] = pd.DataFrame(np.array(par), index=sample_names, columns=contexts)
            elif (parname == "alpha_prior" or parname == "alpha_prior_unn"): 
                param_dict[parname] = pd.DataFrame(np.array(par), index=range(self.n_groups), columns=fixed_names+denovo_names)
            elif parname == "alpha_star": 
                param_dict[parname] = pd.DataFrame(np.array(par), index=sample_names, columns=denovo_names)
            elif parname == "post_probs" and isinstance(par, torch.Tensor):
                param_dict[parname] = pd.DataFrame(np.array(torch.transpose(par, dim0=1, dim1=0)), index=sample_names , columns=range(self.n_groups))
        return param_dict


    def _set_init_params(self, sample_names, fixed_names, denovo_names, contexts, alpha_columns):
        # return
        for k, v_tmp in self.init_params.items():
            v = self._to_cpu(v_tmp, move=True)
            if v is None: continue

            if k == "alpha":
                self.init_params[k] = pd.DataFrame(np.array(v), index=sample_names, columns=alpha_columns)
            elif k == "beta_dn_param":
                self.init_params[k] = pd.DataFrame(np.array(v), index=denovo_names, columns=contexts)
            elif k == "alpha_prior_param":
                self.init_params[k] = pd.DataFrame(np.array(v), index=range(self.n_groups), columns=fixed_names+denovo_names)
            elif k == "beta_weights":
                self.init_params[k] = pd.DataFrame(np.array(v), index=denovo_names, columns=fixed_names+["DN"])
            elif k == "epsilon_var":
                self.init_params[k] = pd.DataFrame(np.array(v), index=sample_names, columns=contexts)
            else:
                self.init_params[k] = np.array(v)




import torch
import pyro
import pyro.distributions.constraints as constraints
import pyro.distributions as dist
import pyro.poutine as poutine
import torch.nn.functional as F
import numpy as np
import pandas as pd

from pyro.infer import SVI, Trace_ELBO
from pyro.infer.autoguide import AutoDelta
from pyro.ops.indexing import Vindex
from sklearn.cluster import KMeans
from scipy.optimize import direct
from collections import defaultdict
from copy import deepcopy
from tqdm import trange

MIN_POS_N = torch.finfo(torch.float32).tiny

class PyBasilica_mixture():
    def __init__(
        self,
        alpha,
        n_steps,
        lr = 0.005,
        optim_gamma = 0.1,
        enumer = "parallel",
        cluster = 5,
        hyperparameters = None,

        compile_model = False,
        CUDA = False,
        store_parameters = False,

        seed = 10,
        nonparam = True,
        autoguide = False
        ):

        self._hyperpars_default = {"pi_conc0":0.6, 
                                   "z_tau":0.1,
                                   "scale_factor_centroid":1, 
                                   "alpha_thr":0.01}
        
        self.autoguide = autoguide
        self._set_data_catalogue(alpha)

        self._set_fit_settings(lr=lr, optim_gamma=optim_gamma, n_steps=n_steps, \
                               compile_model=compile_model, CUDA=CUDA, store_parameters=store_parameters,
                               seed=seed, nonparam=nonparam)
        self._set_hyperparams(enumer=enumer, cluster=cluster, hyperparameters=hyperparameters)
        self._init_seed = 15


    def _set_fit_settings(self, lr, optim_gamma, n_steps, compile_model, CUDA, \
                          store_parameters, seed, nonparam):
        self.lr = lr
        self.optim_gamma = optim_gamma
        self.n_steps = int(n_steps)
        self.compile_model = compile_model
        self.CUDA = CUDA

        self.store_parameters = store_parameters

        self.seed = seed
        self.nonparam = nonparam


    def _set_hyperparams(self, enumer, cluster, hyperparameters):
        self.enumer = enumer

        self.cluster = int(cluster)
        self.n_groups = self.cluster

        self.init_params = None

        if hyperparameters is None:
            self.hyperparameters = self._hyperpars_default
        else:
            self.hyperparameters = {}
            for parname in self._hyperpars_default.keys():
                self.hyperparameters[parname] = hyperparameters.get(parname, self._hyperpars_default[parname])


    def _pad_tensor(self, param):
        '''
        `param`: list of tensors for which the rightmost dimension will be padded \\
        output: list of matrices of same dimension N x K
        '''
        max_shape = max([i.shape[-1] for i in param])
        for i, v in enumerate(param):
            if v.shape[-1] < max_shape:
                pad_dims = max_shape - v.shape[-1]
                v = F.pad(input=v, pad=(0, pad_dims, 0, 0), mode="constant", value=MIN_POS_N)
                param[i] = self._norm_and_clamp(v)
        return param


    def _set_data_catalogue(self, alpha) -> torch.Tensor:
        '''
        `alpha`: either a list (len=V) of N x K matrices or a single N x K matrix (if V=1) \\
        each matrix will be padded if necessary (adding missing K=0) and converted to tensor \\
        output: a tensor of shape N x V x K
        '''
        if not isinstance(alpha, list): alpha = [alpha]

        self.variants_shapes = [i.shape[1] for i in alpha]

        alpha_stacked = deepcopy(alpha)
        for i, v in enumerate(alpha_stacked):  # for each alpha matrix
            if isinstance(v, pd.DataFrame):
                v = torch.tensor(v.values, dtype=torch.float64)
            if not isinstance(v, torch.Tensor):
                v = torch.tensor(v, dtype=torch.float64)
            alpha_stacked[i] = v

        # alpha_stacked -> V x N x K, self.alpha -> N x V x K
        self.alpha = torch.permute(torch.stack(self._pad_tensor(alpha_stacked)), (1, 0, 2))

        self.n_samples = self.alpha.shape[0]
        self.K = self.alpha.shape[-1]
        self.n_variants = self.alpha.shape[-2]


    def _mix_weights(self, beta):
        '''
        Function used for the stick-breaking process.
        '''
        beta1m_cumprod = (1 - beta).cumprod(-1)
        return F.pad(beta, (0, 1), value=1) * F.pad(beta1m_cumprod, (1, 0), value=1)


    # alpha -> N x V x K; alpha_prior -> G x V x K; pi -> [G]
    def model_mixture(self):
        cluster, n_samples, n_variants = self.cluster, self.n_samples, self.n_variants
        pi_conc0 = self.hyperparameters["pi_conc0"]
        z_tau = torch.tensor(self.hyperparameters["z_tau"]).double()

        if self.nonparam:
            with pyro.plate("beta_plate", cluster-1):
                pi_beta = pyro.sample("beta_pi", dist.Beta(1, pi_conc0))
            pi = self._mix_weights(pi_beta)
        else:
            pi = pyro.sample("pi", dist.Dirichlet(self.init_params["pi"]))

        # alpha_prior = V x G x K 
        sf_conc = self.init_params["scale_factor_centroid"]
        alpha_centr = torch.permute(self.init_params["alpha_prior"], (1,0,2))
        with pyro.plate("g1", cluster, dim=-1):
            with pyro.plate("n_vars1", n_variants, dim=-2):
                sf_centroid = pyro.sample("scale_factor_centroid", dist.Gamma(sf_conc*2, 2))
                alpha_prior = pyro.sample("alpha_prior", dist.Dirichlet(alpha_centr * sf_centroid.unsqueeze(-1)))

        tau = pyro.sample("z_tau", dist.Normal(z_tau, 10.))
        # alpha = N x V x K
        with pyro.plate("n", n_samples):
            z_onehot = pyro.sample("latent_class", dist.RelaxedOneHotCategorical(temperature=tau, 
                                                                                logits=torch.log(pi)))
            z = z_onehot.argmax(dim=-1).long()
            for v in pyro.plate("n_vars2", n_variants):
                # alpha -> V x N x K
                # alpha_prior -> V x G x K
                alpha_c = Vindex(alpha_prior)[v,z,:]
                pyro.sample(f"obs_{v}", dist.Dirichlet(alpha_c), obs=self.alpha[:,v,:])


    def autoguide_mixture(self):
        self._initialize_params()
        if self.nonparam:
            return AutoDelta(poutine.block(self.model_mixture, 
                                        expose=["scale_factor_centroid", "beta_pi", "alpha_prior", 
                                                "z_tau", "latent_class"]), 
                            init_loc_fn=self.init_fn)
        return AutoDelta(poutine.block(self.model_mixture, 
                                        expose=["scale_factor_centroid", "pi", "alpha_prior", 
                                                "z_tau", "latent_class"]), 
                            init_loc_fn=self.init_fn)


    def init_fn(self, site):
        site_name = site["name"]
        if site_name == "alpha_prior": 
            return torch.permute(self.init_params["alpha_prior"], (1,0,2))
        if site_name == "z_tau":
            return torch.tensor(self.hyperparameters["z_tau"]).double()
        return self.init_params[site_name]


    def guide_mixture(self):
        def guide():
            cluster, n_samples, n_variants = self.cluster, self.n_samples, self.n_variants
            init_params = self._initialize_params()
            z_tau = torch.tensor(self.hyperparameters["z_tau"]).double()
            init_pi = init_params["pi"]

            if self.nonparam:
                pi_param = pyro.param("pi_param", lambda: init_pi, constraint=constraints.simplex)
                pi_conc0 = pyro.param("pi_conc0_param", lambda: (init_pi[:-1])*10,
                                    constraint=constraints.greater_than_eq(MIN_POS_N))
                with pyro.plate("beta_plate", cluster-1):
                    pyro.sample("beta_pi", dist.Beta(1, pi_conc0))
            else:
                pi_param = pyro.param("pi_param", lambda: init_pi, constraint=constraints.simplex)
                pyro.sample("pi", dist.Delta(pi_param).to_event(1))

            sf_centroid_param = pyro.param("scale_factor_centroid_param", lambda: init_params["scale_factor_centroid"])

            init_alpha = torch.permute(init_params["alpha_prior"], (1,0,2))
            alpha_prior_param = pyro.param("alpha_prior_param", lambda: init_alpha, constraint=constraints.simplex)
            with pyro.plate("g1", cluster):
                with pyro.plate("n_vars1", n_variants):
                    pyro.sample("scale_factor_centroid", dist.Delta(sf_centroid_param))
                    pyro.sample("alpha_prior", dist.Delta(alpha_prior_param).to_event(1))

            tau = pyro.param("z_tau_param", z_tau)
            pyro.sample("z_tau", dist.Delta(tau))
            with pyro.plate("n", n_samples):
                pyro.sample("latent_class", dist.RelaxedOneHotCategorical(temperature=tau, 
                                                                        logits=torch.log(pi_param)))

        return guide


    def _check_kmeans(self, X):
        '''
        If there are less than 2 identical samples in the data the kmeans cannot be run.
        '''
        if X.unique(dim=0).shape[0] < 2:
            return False
        return True


    def run_kmeans(self, X, G, seed):
        X = self._to_cpu(X, move=True)
        try:
            km = KMeans(n_clusters=G, random_state=seed, n_init=10).fit(X.numpy())

        except:
            removed_idx, data_unq = self.check_input_kmeans(X.numpy())  # if multiple points are equal the function will throw an error
            km = KMeans(n_clusters=G, random_state=seed, n_init=10).fit(data_unq)

            clusters = km.labels_
            for rm in sorted(removed_idx.keys()):
                clusters = np.insert(clusters, rm, 0, 0)  # insert 0 elements to restore the original number of obs

            for rm in removed_idx.keys():
                rpt = removed_idx[rm]  # the index of the kept row
                clusters[rm] = clusters[rpt]  # insert in the repeated elements the correct cluster
        return km


    def check_input_kmeans(self, counts):
        '''
        Function to check the inputs of the Kmeans. There might be a problem when multiple observations
        are equal since the Kmeans will keep only a unique copy of each and the others will not be initialized.
        '''
        tmp, indexes, count = np.unique(counts, axis=0, return_counts=True, return_index=True)
        repeated_groups = tmp[count > 1].tolist()

        unq = np.array([counts[index] for index in sorted(indexes)])

        removed_idx = {}
        for i, repeated_group in enumerate(repeated_groups):
            rpt_idxs = np.argwhere(np.all(counts == repeated_group, axis=1)).flatten()
            removed = rpt_idxs[1:]
            for rm in removed:
                removed_idx[rm] = rpt_idxs[0]

        return removed_idx, unq


    def _initialize_params_clustering(self):
        '''
        Function to initialize the values for each parameter.
        The pars are initialized through a Kmeans, run on the optimal number of clusters G.
        `alpha` is permuted and concatenated to have shape N x K*V
        `alpha_prior` is estimated as G x K*V
        '''
        pi = alpha_prior = None

        alpha_perm = torch.permute(self.alpha.clone(), (1,0,2))
        alpha_km = torch.cat(tuple(alpha_perm), dim=1)

        pi_km = torch.ones((self.cluster,))
        alpha_prior_km = torch.ones((self.cluster, self.K * self.n_variants)) / self.K
        
        km = self.run_kmeans(X=alpha_km, G=self.cluster, seed=self._init_seed)
        clusters = torch.tensor(km.labels_)

        ## setup mixing proportion vector
        pi_km = torch.tensor([(torch.where(clusters == g)[0].shape[0]) for g in range(km.n_clusters)], dtype=torch.float64)
        pi_km = dist.Dirichlet(pi_km).sample()
        pi_km[pi_km < 1e-15] = 1e-15
        pi_km = self._norm_and_clamp(pi_km)

        alpha_prior_km = torch.tensor(km.cluster_centers_)
        last, alpha_prior = 0, list()
        for _ in range(self.n_variants):
            alpha_p_tmp = dist.Dirichlet(alpha_prior_km[:, last:(last + self.K)]).sample()
            alpha_p_tmp[alpha_p_tmp < MIN_POS_N] = MIN_POS_N
            alpha_p_tmp = self._norm_and_clamp(alpha_p_tmp)

            alpha_prior.append(alpha_p_tmp)
            last = last + self.K

        latent_class_km = dist.RelaxedOneHotCategorical(temperature=torch.tensor(0.1), 
                                                        logits=torch.log(pi_km)).sample((self.n_samples,))
        latent_class_km[latent_class_km < MIN_POS_N] = MIN_POS_N

        latent_class =  self._to_gpu(latent_class_km, move=True)
        pi = self._to_gpu(pi_km, move=True)
        alpha_prior = self._to_gpu(torch.stack(alpha_prior), move=True)
        alpha_prior = torch.permute(alpha_prior.clone().detach().double(), (1,0,2))
        variances = self._compute_empirical_variance(alpha=self.alpha, 
                                                     clusters=clusters)
        init_scale_factors = self._compute_scale_factors(alpha_prior=alpha_prior, variances=variances)
        sf_centroid = init_scale_factors.repeat(self.cluster, 1).permute((1,0)).double()

        return self._set_init_params_dict(alpha_prior=alpha_prior, 
                                          pi=pi.clone().detach().double(), 
                                          latent_class=latent_class.clone().detach().double(),
                                          variances=variances,
                                          sf_centroid=sf_centroid,
                                          clusters=clusters)


    def _compute_empirical_variance(self, alpha, clusters):
        variances = torch.zeros(self.cluster, self.n_variants, self.K)
        for g in range(self.cluster):
            alpha_g = alpha[torch.where(clusters == g)[0]]
            for v in range(self.n_variants):
                if alpha_g.shape[0] > 0:
                    alpha_v = alpha_g[:,v,:]
                    variances[g, v] = torch.var(alpha_v, dim=0)

        variances[variances < MIN_POS_N] = MIN_POS_N
        assert variances.shape == (self.cluster, self.n_variants, self.K)
        return variances


    def _compute_scale_factors(self, variances=None, alpha_prior=None, convert=False, to_cpu=False):
        if variances is None: variances = self.init_params["variances"]
        if variances is None: return None
        if alpha_prior is None:
            alpha_prior = self.get_param_dict(convert=False, permute=True, concat=False, to_cpu=False)["alpha_prior"]

        variances = self._to_cpu(variances)
        alpha_prior = self._to_cpu(alpha_prior)

        assert variances.shape == (self.cluster, self.n_variants, self.K)
        assert alpha_prior.shape == (self.cluster, self.n_variants, self.K)

        sf_vector = torch.zeros(self.cluster, self.n_variants).double()
        for g in range(self.cluster):
            for v in range(self.n_variants):
                # variances is a GxVxK tensor with variances in alpha values for each group/variant/signature
                # for each group/variant we test several values of `sf` (bounds=[1,1000])
                # `direct` will minimize the function provided in `func`
                # we compare the empirical variance of the alpha matrix with the variance of 
                # the Dirichlet with parameters the rescaled centroids [higher SF -> lower variance]
                # we're minimising the distance between empirical and theoretical variance
                sf_gv = direct(func=self.variance_condition, bounds=([1.,1000.],), 
                               args=(variances[g, v], alpha_prior[g, v]), maxiter=100).x
                if isinstance(sf_gv, torch.Tensor): sf_vector[g, v] = sf_gv.clone().detach()
                else: sf_vector[g, v] = self._to_gpu(torch.tensor(sf_gv))
        
        # keep the min for each variant
        scale_factors = torch.amax(sf_vector, dim=0)
        scale_factors[scale_factors < 1.] = 1.
        scale_factors = self._convert(scale_factors, convert=convert, to_cpu=to_cpu)
        assert scale_factors.shape == (self.n_variants,)
        return scale_factors


    def variance_condition(self, sf, target, alpha):
        variances = self._compute_dirichlet_variance(alpha * sf)
        return torch.amax(torch.abs(variances - target))


    def _compute_dirichlet_variance(self, alpha):
        alpha_hat = torch.sum(alpha)
        num = alpha * (alpha_hat - alpha)
        denomin = alpha_hat**2 * (alpha_hat + 1)
        return num / denomin


    def _set_init_params_dict(self, alpha_prior=None, pi=None, latent_class=None, variances=None, 
                              sf_centroid=None, clusters=None):
        return {"alpha_prior":alpha_prior,
                "pi":pi,
                "beta_pi":pi[:-1],
                "scale_factor_centroid":sf_centroid,
                "latent_class":latent_class,
                "variances":variances,
                "init_clusters":clusters}


    def _initialize_params_random(self):
        pi = dist.Dirichlet(torch.ones(self.cluster, dtype=torch.float64)).sample()
        alpha_prior = dist.Dirichlet(torch.ones(self.cluster, self.n_variants, self.K, dtype=torch.float64)).sample()
        latent_class = dist.RelaxedOneHotCategorical(temperature=torch.tensor(0.1), 
                                                                   logits=torch.log(pi)).sample((self.n_samples,))
        latent_class[latent_class < MIN_POS_N] = MIN_POS_N
        sf_centroid = torch.ones(self.n_variants, self.cluster) * self.hyperparameters["scale_factor_centroid"]

        return self._set_init_params_dict(alpha_prior=alpha_prior, pi=pi, 
                                          sf_centroid=sf_centroid, latent_class=latent_class)


    def _initialize_params(self):
        if self.init_params is None:
            if self._check_kmeans(torch.cat(tuple(self.alpha.clone()), dim=0)) and self.cluster > 1:
                self.init_params = self._initialize_params_clustering()
            else:
                self.init_params = self._initialize_params_random()
        return self.init_params


    def _fit(self):
        pyro.clear_param_store()  # always clear the store before the inference

        if self.CUDA and torch.cuda.is_available():
            torch.set_default_tensor_type("torch.cuda.DoubleTensor")
        else:
            torch.set_default_tensor_type(t=torch.DoubleTensor)

        pyro.set_rng_seed(self.seed)
        pyro.get_param_store().clear()

        self.alpha = self._to_gpu(self.alpha)

        self._initialize_params()

        elbo = Trace_ELBO()
        optimizer = pyro.optim.Adam({"lr": self.lr})

        guide = self.autoguide_mixture() if self.autoguide else self.guide_mixture()
        svi = SVI(self.model_mixture, guide, optimizer, loss=elbo)
        loss = float(svi.step())

        train_params_each = 2 if self.n_steps <= 100 else int(self.n_steps / 100)
        gradient_norms = defaultdict(list)
        for name, value in pyro.get_param_store().named_parameters():
            value.register_hook(lambda g, name=name: gradient_norms[name].append(g.norm().item()))

        self.losses, self.regs, self.likelihoods, self.train_params = list(), list(), list(), list()
        t = trange(self.n_steps, desc="Bar desc", leave=True)
        for i in t:   # inference - do gradient steps
            loss = float(svi.step())
            self.losses.append(loss)

            # self.likelihoods.append(self._likelihood_mixture(to_cpu=False).sum())
            
            if self.store_parameters and i%train_params_each==0:
                self.train_params.append(self.get_param_dict(convert=True, to_cpu=True))

            if i%5 == 0:
                t.set_description("ELBO %f" % loss)
                t.refresh()

        self.alpha = self._to_cpu(self.alpha)

        self.gradient_norms = dict(gradient_norms) if gradient_norms is not None else {}
        self.params = {**self._set_clusters(to_cpu=True), 
                       **self.get_param_dict(convert=True, permute=False, concat=True, to_cpu=True)}

        self.bic = self.aic = self.icl = self.reg_likelihood = None
        self.likelihood = self._likelihood_mixture(to_cpu=True).sum().item()
        self.set_scores()

        self.init_params = self._get_init_param_dict(convert=True, to_cpu=True)


    def get_param_dict(self, convert=False, permute=False, to_cpu=True, concat=False):
        params = dict()

        if self.autoguide:
            params["alpha_prior"] = self._get_param("AutoDelta.alpha_prior", convert=convert, 
                                                    permute=permute, concat=concat, to_cpu=to_cpu)
            if self.nonparam:
                beta_pi = self._get_param("AutoDelta.beta_pi", convert=False, to_cpu=to_cpu)
                params["pi"] = self._mix_weights(beta_pi)
                if convert: 
                    params["pi"] = self._to_cpu(params["pi"], move=True).numpy()
            else:
                params["pi"] = self._get_param("AutoDelta.pi", convert=convert, to_cpu=to_cpu)
            params["z_tau"] = self._get_param("AutoDelta.z_tau", convert=convert, to_cpu=to_cpu)

            params["scale_factor_centroid"] = self._get_param("AutoDelta.scale_factor_centroid", convert=convert, to_cpu=to_cpu)

        else:
            params["alpha_prior"] = self._get_param("alpha_prior_param", convert=convert, 
                                                    permute=permute, concat=concat, to_cpu=to_cpu)
            params["pi"] = self._get_param("pi_param", convert=convert, to_cpu=to_cpu)
            
            params["scale_factor_centroid"] = self._get_param("scale_factor_centroid_param", convert=convert, to_cpu=to_cpu)
            params["z_tau"] = self._get_param("z_tau_param", convert=convert, to_cpu=to_cpu)

        return params


    def _get_init_param_dict(self, convert=True, to_cpu=True):
        params = dict()
        for k, v in self.init_params.items():
            if v is None: params[k] = v
            elif not convert: params[k] = v
            else:
                v = self._convert(v, convert=convert, to_cpu=to_cpu, permute=True, concat=True)
                params[k] = v
        return params


    def set_scores(self):
        self._set_bic()
        self._set_aic()
        self._set_icl()


    def _get_param(self, param_name, to_cpu=True, permute=False, convert=False, concat=False):
        try:
            par = pyro.param(param_name)
            par = self._convert(par, convert=convert, permute=permute, concat=concat, to_cpu=to_cpu)
            return par

        except Exception as e: return None


    def _convert(self, par, convert=False, permute=False, concat=False, to_cpu=False):
        if par is None: return par
        par = self._to_cpu(par, move=to_cpu)
        if isinstance(par, torch.Tensor): 
            par = par.clone().detach()
            if permute and len(par.shape) > 2: par = torch.permute(par, (1,0,2))
        if concat: par = self._concat_tensors(par, dim=1)
        if convert: 
            par = self._to_cpu(par, move=True)
            par = np.array(par)
        return par


    def _likelihood_mixture(self, to_cpu=False, params=None):
        if params is None: params = self.get_param_dict(convert=False, permute=True, 
                                                        concat=False, to_cpu=to_cpu)
        alpha = self._to_cpu(params.get("alpha", self.alpha.clone()), move=to_cpu)
        alpha_prior = params["alpha_prior"]

        sf = torch.ones(self.n_variants, self.cluster) if params["scale_factor_centroid"] is None else params["scale_factor_centroid"]
        assert sf.shape == (self.n_variants, self.cluster)
        assert alpha_prior.shape == (self.cluster, self.n_variants, self.K)
        assert alpha.shape == (self.n_samples, self.n_variants, self.K)

        thr = self.hyperparameters["alpha_thr"]
        alpha[alpha < thr] = thr
        alpha_prior[alpha_prior < thr] = thr
        alpha = self._norm_and_clamp(alpha)
        alpha_prior = self._norm_and_clamp(alpha_prior)

        llik = torch.zeros(self.cluster, self.n_samples)
        for g in range(self.cluster):
            lprob_alpha = torch.zeros((self.n_variants, self.n_samples))
            for v in range(self.n_variants):
                sf_v = sf[v, g]
                alpha_prior_g = alpha_prior[g, v]

                assert lprob_alpha.shape == (self.n_variants, self.n_samples)

                lprob_alpha[v] = dist.Dirichlet(alpha_prior_g * sf_v).log_prob(alpha[:,v,:])

            # sum over independent variants and add log(pi)
            llik[g, :] = torch.sum(lprob_alpha, dim=0) + torch.log(params["pi"][g])

        assert llik.shape == (self.cluster, self.n_samples)

        return llik


    def _compute_posterior_probs(self, to_cpu=True, compute_exp=True):
        ll_k = self._likelihood_mixture(to_cpu=to_cpu)

        assert ll_k.shape == (self.cluster, self.n_samples)
        
        ll = self._logsumexp(ll_k).unsqueeze(0)

        probs = torch.exp(ll_k - ll) if compute_exp else ll_k - ll
        z = torch.argmax(probs, dim=0)
        return self._to_cpu(z.long(), move=True).numpy(), self._to_cpu(probs, move=to_cpu)


    def _logsumexp(self, weighted_lp) -> torch.Tensor:
        '''
        Returns `m + log( sum( exp( weighted_lp - m ) ) )`
        - `m` is the the maximum value of weighted_lp for each observation among the K values
        - `torch.exp(weighted_lp - m)` to perform some sort of normalization
        In this way the `exp` for the maximum value will be exp(0)=1, while for the
        others will be lower than 1, thus the sum across the K components will sum up to 1.
        '''
        m = torch.amax(weighted_lp, dim=0).unsqueeze(0)  # the maximum value for each observation among the K values
        summed_lk = m[-1] + torch.log(torch.sum(torch.exp(weighted_lp - m), axis=0))
        return summed_lk


    def _norm_and_clamp(self, par):
        mmin = 0
        if torch.any(par < 0): mmin = torch.min(par, dim=-1)[0].unsqueeze(-1)

        nnum = par - mmin
        par = nnum / torch.sum(nnum, dim=-1).unsqueeze(-1)

        return par


    def _set_clusters(self, to_cpu=True):
        params = dict()
        self.groups, params["post_probs"] = self._compute_posterior_probs(to_cpu=to_cpu)
        return params


    def _set_bic(self):
        _log_like = self.likelihood

        k = self._number_of_params()
        n = self.n_samples
        bic = k * torch.log(torch.tensor(n, dtype=torch.float64)) - (2 * _log_like)

        self.bic = bic.item()


    def _set_aic(self):
        _log_like = self.likelihood

        k = self._number_of_params()
        aic = 2*k - 2*_log_like

        if (isinstance(aic, torch.Tensor)):
            self.aic = aic.item()
        else:
            self.aic = aic


    def _set_icl(self):
        self.icl = None
        if self.cluster is not None:
            icl = self.bic + self._compute_entropy()
            self.icl = icl.item()


    def _compute_entropy(self) -> np.array:
        '''
        `entropy(z) = - sum^K( sum^N( z_probs_nk * log(z_probs_nk) ) )`
        `entropy(z) = - sum^K( sum^N( exp(log(z_probs_nk)) * log(z_probs_nk) ) )`
        '''

        logprobs = self._compute_posterior_probs(to_cpu=True, compute_exp=False)[1]
        entr = 0
        for n in range(self.n_samples):
            for k in range(self.cluster):
                entr += torch.exp(logprobs[k,n]) * logprobs[k,n]
        return -entr.detach()


    def _number_of_params(self):
        n_grps = self.cluster
        k = n_grps + self.params["alpha_prior"].shape[1] * n_grps
        return k


    def _concat_tensors(self, param, dim=1):
        if len(param.shape) <= 2: return param
        if isinstance(param, torch.Tensor):
            return torch.cat(tuple(param), dim=dim)
        return np.concatenate(tuple(param), axis=dim)


    def _pad_dataframes(self, param):
        new_param = []
        max_shape = max([i.shape[1] for i in param])
        for i in range(len(param)):
            par_i = param[i]
            if par_i.shape[1] < max_shape:
                pad_dims = max_shape - par_i.shape[1]
                columns = list(str(i)+"_"+par_i.columns) + [str(i)+"_P"+str(j) for j in range(pad_dims)]
                index = par_i.index
                values = F.pad(input=self._to_cpu(torch.tensor(par_i.values), move=True), \
                               pad=(0, pad_dims, 0,0), mode="constant", value=MIN_POS_N)
                new_param.append(pd.DataFrame(values.numpy(), index=index, columns=columns))
            else:
                new_param.append(par_i.copy(deep=True))
                new_param[i].columns = list(str(i)+"_"+par_i.columns)
        return new_param


    def convert_to_dataframe(self, alpha):
        # mutations catalogue
        if isinstance(alpha, list): alpha = [self._to_cpu(i, move=True) for i in alpha]
        if not isinstance(alpha, list): alpha = [self._to_cpu(alpha, move=True)]

        self.alpha = pd.concat(self._pad_dataframes(alpha), axis=1)
        sample_names, sigs = list(self.alpha.index), list(self.alpha.columns)

        if self.groups is not None: self.groups = np.array(self.groups)

        self.params = self._convert_pars(param_dict=self.params, sigs=sigs, sample_names=sample_names)
        if len(self.train_params) > 0:
            for i, v in enumerate(self.train_params):
                self.train_params[i] = self._convert_pars(param_dict=v, sigs=sigs,
                                                          sample_names=sample_names)

        for k, v in self.hyperparameters.items():
            if v is None: continue
            v = self._to_cpu(v, move=True)
            if isinstance(v, torch.Tensor):
                if len(v.shape) == 0: self.hyperparameters[k] = int(v)
                else: self.hyperparameters[k] = v.numpy()

        self._set_init_params(sigs)

        self.n_groups = len(np.unique(self.groups))


    def _convert_pars(self, param_dict, sigs, sample_names):
        for parname, par in param_dict.items():
            if par is None: continue
            par = self._to_cpu(par, move=True)
            if parname == "pi":
                param_dict[parname] = par.tolist() if isinstance(par, torch.Tensor) else par
            elif (parname == "alpha_prior" or parname == "alpha_prior_unn"):
                param_dict[parname] = pd.DataFrame(par, index=range(self.n_groups), columns=sigs)
            elif parname == "scale_factor_matrix":
                param_dict[parname] = np.array(par)
            elif parname == "post_probs" and isinstance(par, torch.Tensor):
                param_dict[parname] = pd.DataFrame(np.array(torch.transpose(par, dim0=1, dim1=0)), index=sample_names , columns=range(self.n_groups))
        return param_dict


    def _set_init_params(self, sigs):
        for k, v_tmp in self.init_params.items():
            v = self._to_cpu(v_tmp, move=True)
            if v is None: continue

            if k == "alpha_prior":
                self.init_params[k] = pd.DataFrame(v, index=range(self.n_groups), columns=sigs)
            elif k == "variances":
                self.init_params[k] = pd.DataFrame(v, columns=sigs)
            else:
                self.init_params[k] = np.array(v)


    def _to_cpu(self, param, move=True):
        if param is None: return None
        if move and self.CUDA and torch.cuda.is_available() and isinstance(param, torch.Tensor):
            return param.cpu()
        return param


    def _to_gpu(self, param, move=True):
        if param is None: return None
        if move and self.CUDA and torch.cuda.is_available() and isinstance(param, torch.Tensor):
            return param.cuda()
        return param





import random
import pyro
import pyro.distributions as dist
import numpy as np
import pandas as pd
import torch

def norm_and_clamp(par):
    mmin = 0
    if torch.any(par < 0): mmin = torch.min(par, dim=-1)[0].unsqueeze(-1)

    nnum = par - mmin
    par = nnum / torch.sum(nnum, dim=-1).unsqueeze(-1)

    return par


def to_cpu(param, move=True):
    if param is None: return None
    if move and torch.cuda.is_available() and isinstance(param, torch.Tensor):
        return param.cpu()
    return param


def generate_model(alpha_prior, beta, n_muts, N, alpha_sigma=0.05, seed=15, use_normal=True):
    input_beta = beta
    if isinstance(beta, pd.DataFrame):
        beta = torch.tensor(beta.values)
    alpha_prior = torch.tensor(alpha_prior, dtype=torch.float64)
    beta = beta.clone().detach().double()
    n_muts = torch.tensor(n_muts, dtype=torch.float64)

    pyro.set_rng_seed(seed)

    alpha_prior = norm_and_clamp(alpha_prior)
    beta = norm_and_clamp(beta)

    q_01 = alpha_prior - alpha_sigma # * alpha_prior
    q_99 = alpha_prior + alpha_sigma # * alpha_prior
    alpha_sigma_corr = (q_99 - q_01) / (2 * dist.Normal(alpha_prior, 1).icdf(torch.tensor(0.99)))  # good clustering

    # Observations
    with pyro.plate("n2", N):
        if use_normal:
            alpha  = pyro.sample("latent_exposure", dist.Normal(alpha_prior, alpha_sigma_corr).to_event(1))  # good clustering
        else:
            alpha  = pyro.sample("latent_exposure", dist.Cauchy(alpha_prior, alpha_sigma_corr).to_event(1))  # good clustering

        alpha = norm_and_clamp(alpha)

        a = torch.matmul(torch.matmul(torch.diag(n_muts), alpha), beta)
        data = pyro.sample("obs", dist.Poisson(a).to_event(1))

    return {"data":pd.DataFrame(to_cpu(data, move=True), columns=input_beta.columns), 
            "alpha":pd.DataFrame(to_cpu(alpha, move=True), columns=input_beta.index), 
            "alpha_sigma":to_cpu(alpha_sigma_corr, move=True), 
            "beta":input_beta}




# #-----------------------------------------------------------------[<QC-PASSED>]
# # divide cosmic signatures into two parts:
# # 1st part: considered as cosmic signatures
# # 2nd part: considered as denovo signatures

# def cosmic_denovo(full_cosmic_df):

#     #random.seed(a=seed)

#     # cosmic_path --- <class 'str'>
#     #full_cosmic_df = pd.read_csv(cosmic_path, index_col=0)
#     full_cosmic_list = list(full_cosmic_df.index)

#     cosmic_list = random.sample(full_cosmic_list, k=50)
#     denovo_list = list(set(full_cosmic_list) - set(cosmic_list))

#     cosmic_df = full_cosmic_df.loc[cosmic_list] # <class 'pandas.core.frame.DataFrame'>
#     denovo_df = full_cosmic_df.loc[denovo_list] # <class 'pandas.core.frame.DataFrame'>

#     return cosmic_df, denovo_df
#     # cosmic_df --- dtype: DataFrame
#     # denovo_df --- dtype: DataFrame

# #-----------------------------------------------------------------[QC:PASSED]

# def target_generator(cosmic_df, denovo_df, target_complexity, num_samples):
#     # profile ------- {"A", "B", "C"}
#     # cosmic_df ----- <class 'pandas.core.frame.DataFrame'>
#     # denovo_df ----- <class 'pandas.core.frame.DataFrame'>

#     #random.seed(a=seed)

#     # num_samples = random.randint(15, 25)
#     num_samples = int(num_samples)

#     # error handling for profile argument
#     valid = ["low", "medium", "high"]
#     if target_complexity not in valid:
#         raise ValueError("profile must be one of %s." % valid)

#     if target_complexity=="low":
#         fixed_num = random.randint(3, 5)    # <class 'int'>
#         denovo_num = random.randint(0, 2)
#     elif target_complexity=="medium":
#         fixed_num = random.randint(0, 2)
#         denovo_num = random.randint(3, 5)
#     elif target_complexity=="high":
#         fixed_num = random.randint(3, 5)
#         denovo_num = random.randint(3, 5)

#     cosmic_list = list(cosmic_df.index)
#     denovo_list = list(denovo_df.index)
#     mutation_features = list(cosmic_df.columns)

#     # beta fixed ------------------------------------------------------
#     if fixed_num > 0:
#         fixed_list = random.sample(cosmic_list, k=fixed_num)
#         beta_fixed_df = cosmic_df.loc[fixed_list]
#     else:
#         beta_fixed_df = pd.DataFrame(columns=mutation_features)
#         #beta_fixed_df = None
    
#     # beta denovo -----------------------------------------------------
#     if denovo_num > 0:
#         denovo_list = random.sample(denovo_list, k=denovo_num)
#         beta_denovo_df = denovo_df.loc[denovo_list]

#         # add "_D" to the end of denovo signatures to distinguish them from fixed signatures
#         denovo_labels = []
#         for i in range(len(denovo_list)):
#             denovo_labels.append(denovo_list[i] + '_D')
#         beta_denovo_df.index = denovo_labels
#     else:
#         beta_denovo_df = pd.DataFrame(columns=mutation_features)
#         #beta_denovo_df = None

#     if beta_denovo_df.empty:
#         beta_df = beta_fixed_df
#     elif beta_fixed_df.empty:
#         beta_df = beta_denovo_df
#     else:
#         beta_df = pd.concat([beta_fixed_df, beta_denovo_df], axis=0)

#     signatures = list(beta_df.index)
#     beta_tensor = torch.tensor(beta_df.values).float()

#     #------- alpha ----------------------------------------------
#     matrix = np.random.rand(num_samples, len(signatures))
#     alpha_tensor = torch.tensor(matrix / matrix.sum(axis=1)[:, None]).float()
#     alpha_np = np.array(alpha_tensor)
#     alpha_df = pd.DataFrame(alpha_np, columns=signatures)

#     #------- theta ----------------------------------------------
#     theta = random.sample(range(1000, 4000), k=num_samples) # dtype:list
    
#     #------- check dimensions -----------------------------------
#     m_alpha = alpha_tensor.size()[0]    # no. of branches (alpha)
#     m_theta = len(theta)                # no. of branches (theta)
#     k_alpha = alpha_tensor.size()[1]    # no. of signatures (alpha)
#     k_beta = beta_tensor.size()[0]      # no. of signatures (beta)
#     if not(m_alpha == m_theta and k_alpha == k_beta):
#         raise ValueError("wrong dimensions!")
    
#     M_tensor = torch.zeros([num_samples, 96])   # initialize mutational catalogue with zeros

#     for i in range(num_samples):
#         p = alpha_tensor[i]         # selecting branch i
#         for k in range(theta[i]):   # iterate for number of the mutations in branch i

#             # sample signature profile index from categorical data
#             b = beta_tensor[dist.Categorical(p).sample().item()]

#             # sample mutation feature index for corresponding signature from categorical data
#             j = dist.Categorical(b).sample().item()

#             # add +1 to the mutation feature in position j in branch i
#             M_tensor[i, j] += 1

#     # phylogeny
#     M_np = np.array(M_tensor)
#     M_df = pd.DataFrame(M_np, columns=mutation_features)
#     M_df = M_df.astype(int)

#     # return all in dataframe format
#     return M_df, alpha_df, beta_fixed_df, beta_denovo_df
#     # M_df ------------- dtype: dataframe
#     # alpha_df --------- dtype: dataframe
#     # beta_fixed_df ---- dtype: dataframe, could be empty (if empty --> beta_denovo_df != empty)
#     # beta_denovo_df --- dtype: dataframe, could be empty (if empty --> beta_fixed_df != empty)

# #-----------------------------------------------------------------[QC:PASSED]
# def input_catalogue_generator(cosmic_df, beta_fixed_df, beta_denovo_df, input_complexity):
#     # profile ---------- {"X", "Y", "Z"}
#     # beta_fixed_df ---- dtype: dataframe
#     # beta_denovo_df --- dtype: dataframe
#     # cosmic_df -------- dtype: dataframe

#     #random.seed(a=seed)

#     # error handling for profile argument
#     valid = ["low", "medium", "high"]
#     if input_complexity not in valid:
#         raise ValueError("profile must be one of %s." % valid)

#     # TARGET DATA -----------------------------------------------------------------------------
#     beta_fixed_names = list(beta_fixed_df.index)    # target beta signatures names (dtype: list)
#     k_fixed_target = len(beta_fixed_names)
#     beta_denovo_names = list(beta_denovo_df.index)  # target beta signatures names (dtype: list)
#     k_denovo_target = len(beta_denovo_names)

#     # common fixed signatures (target intersect test)
#     # different fixed signatures (test minus target)
#     if input_complexity=="low":
#         if k_fixed_target > 0:
#             k_overlap = random.randint(1, k_fixed_target)
#             k_extra = 0
#         else:
#             k_overlap = 0
#             k_extra = random.randint(1, k_denovo_target)

#     elif input_complexity=="medium":
#         if k_fixed_target > 0:
#             k_overlap = random.randint(1, k_fixed_target)
#             k_extra = random.randint(1, k_fixed_target)
#         else:
#             k_overlap = 0
#             k_extra = random.randint(1, k_denovo_target)

#     elif input_complexity=="high":
#         k_overlap = 0
#         k_extra = random.randint(1, k_fixed_target + k_denovo_target)
    
    
#     cosmic_names = list(cosmic_df.index)    # cosmic signatures names (dtype: list)

#     # exclude beta target fixed signatures from cosmic
#     for signature in beta_fixed_names:
#         cosmic_names.remove(signature)
    
#     # common fixed signatures list
#     if k_overlap > 0:
#         overlap_sigs = random.sample(beta_fixed_names, k=k_overlap)
#     else:
#         overlap_sigs = []
    
#     # different fixed signatures list
#     if k_extra > 0:
#         extra_sigs = random.sample(cosmic_names, k=k_extra)
#     else:
#         extra_sigs = []
    
#     beta_input = cosmic_df.loc[overlap_sigs + extra_sigs]

#     return beta_input   # dtype: dataframe


# #-----------------------------------------------------------------[PASSED]
# def input_generator(full_cosmic_df, target_complexity, input_complexity, num_samples):

#     cosmic_df, denovo_df = cosmic_denovo(full_cosmic_df)
    
#     M_df, alpha_df, beta_fixed_df, beta_denovo_df = target_generator(cosmic_df, denovo_df, target_complexity, num_samples)
    
#     beta_input_df = input_catalogue_generator(cosmic_df, beta_fixed_df, beta_denovo_df, input_complexity)
    
#     #beta_df = pd.concat([beta_fixed_df, beta_denovo_df], axis=0)

#     if beta_fixed_df.empty:
#         beta_fixed_df = None

#     if beta_denovo_df.empty:
#         beta_denovo_df = None

#     if beta_input_df.empty:
#         beta_input_df = None

#     data = {
#         "M" : M_df,                         # dataframe
#         "alpha" : alpha_df,                 # dataframe
#         "beta_fixed" : beta_fixed_df,       # dataframe
#         "beta_denovo" : beta_denovo_df,     # dataframe
#         "beta_input" : beta_input_df,       # dataframe
#         "cosmic_df" : cosmic_df             # dataframe
#     }

#     return data





import warnings
from setuptools import setup
from setuptools.command.install import install

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

class CustomInstallCommand(install):
    def run(self):
        warnings.simplefilter('default')
        warnings.warn(
            "You are installing 'your_package', which is deprecated. "
            "Please use 'new_package_name' instead.",
            DeprecationWarning
        )
        install.run(self)

setup(
    name = "pybasilica",
    version = "0.4.0",
    description = "Bayesian NMF signatures deconvolution and DP clustering.",
    long_description = long_description,
    long_description_content_type = "text/markdown",
    url = "https://github.com/caravagnalab/pybasilica",
    packages = ["pybasilica"],
    python_requires = ">=3.8",
    install_requires = [
        "pandas>=1.4.2",
        "pyro-api==0.1.2",
        "pyro-ppl==1.8.0",
        "numpy>=1.21.5",
        "torch==1.*",
        "tqdm",
        "rich",
        "statsmodels",
        "uniplot"
        ],
    cmdclass={
        "install": CustomInstallCommand,
    },
)
